#include <string.h>
#include <stdarg.h>

struct str
  {
    char	*s;
    size_t	len;
  };

static struct str *
str_setlen(struct str *p, size_t len)
{
  p->s		= re_alloc(p->s, len);
  p->len	= len;
  return p;
}

static struct str *
str_ext(struct str *p)
{
  return str_setlen(p, p->len+BUFSIZ);
}

static struct str *
str_def(struct str *p)
{
  if (!p->s)
    {
      p->len	= 0;
      str_ext(p);
    }
  return p;
}

static struct str *
str_reset(struct str *p)
{
  DP(("str_reset(%p)", p));
  str_def(p);
  p->s[0]	= 0;
  return p;
}

static struct str *
str_init(struct str *p)
{
  p->s	= 0;
  return str_def(p);
}

static struct str *
str_free(struct str *p)
{
  if (p->s)
    free(p->s);
  p->s		= 0;
  p->len	= 0;
  return p;
}

static struct str *
str_gen(void)
{
  return alloc(sizeof (struct str));
}

static struct str *
str_new(void)
{
  return str_init(str_gen());
}

static void
str_release(struct str *p)
{
  str_free(p);
  free(p);
}

#define STR_INC(P,L)	do { if ((L)>=(P)->len) { str_ext((P)); if ((L)>=(P)->len) str_setlen((P), (L)); } } while (0)

static struct str *
str_inc(struct str *p, size_t len)
{
  STR_INC(p, len);
  if (len>=p->len)
    {
      str_ext(p);
      if (len>=p->len)
	str_setlen(p, len+1);
    }
  return p;
}

static struct str *
str_store(struct str *p, size_t is, const char *s, size_t plus)
{
  size_t	len;

  len	= is+plus;
  str_inc(p, len);
  memcpy(p->s+is, s, plus);
  p->s[len]	= 0;
  return p;
}

static struct str *
str_append_n(struct str *p, const char *s, size_t plus)
{
  return str_store(p, strlen(p->s), s, plus);
}

static struct str *
str_append(struct str *p, const char *s)
{
  return str_append_n(p, s, strlen(s));
}

static struct str *
str_dup(struct str *p)
{
  size_t	len;

  len	= strlen(p->s);
  return str_store(str_setlen(str_gen(), len+1), (size_t)0, p->s, len);
}

#define STR_ADDC(P,I,C)	do { STR_INC((P),(I)); (P)->s[(I)]=(C); } while (0)

static struct str *
str_addc(struct str *p, int index, char c)
{
  xDP(("str_addc(%p, %d, '%c')", p, index, c));
  STR_ADDC(p,index,c);
  return p;
}

static struct str *
str_getline(struct str *p, FILE *fd)
{
  int	i;
  int	c;

  DP(("str_getline(%p, %p)", p, fd));
  str_reset(p);
  for (i=0; ((c=getc(fd))!=EOF && c!='\n'); i++)
    str_addc(p, i, c);
  str_addc(p, i, 0);

  DP(("str_getline() exit"));
  return p;
}

static struct str *
str_fv(struct str *p, const char *s, va_list list)
{
  DP(("str_fv(%p,%s,%p", p,s,list));
  for (;;)
    {
      int	n;

      DP(("str_fv p->s=%p p->len=%d", p->s, p->len));
      n	= vsnprintf(p->s, p->len, s, list);
      DP(("str_fv print=%d", n));
      if (n>=0 && n<p->len-1)
	break;
      str_ext(p);
    }
  DP(("str_fv %p", p));
  return p;
}

static struct str *
str_f(struct str *p, const char *s, ...)
{
  va_list	list;

  DP(("str_f(%p,%s,...", p,s));
  va_start(list, s);
  str_fv(p, s, list);
  va_end(list);
  return p;
}

